/*
 * Copyright (c) 2004-2007 Erez Zadok
 * Copyright (c) 2004-2007 Stony Brook University
 * Copyright (c) 2004-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include "fist.h"
#include "prefetch.h"

#define TIME_BASED_WINDOW_SIZE 1
#define WINDOW_TIME_SIZE 1000 //In milliseconds
#define WINDOW_CMD_SIZE 100   //Number of commands

//Turn off Dynamic Window Sizes
#define WINDOW_SHRINK 2 //Linear - subtract by this number
#define WINDOW_EXPAND 2 //Exponential - multiply by this number

static int prefetch_window_size;
static int last_rcmd_num = 0;

int prefetch_init()
{
  int err = 0;

  print_entry_location();

  if(TIME_BASED_WINDOW_SIZE)
    prefetch_window_size = WINDOW_TIME_SIZE;
  else
    prefetch_window_size = WINDOW_CMD_SIZE;
  last_rcmd_num = 0;

  print_exit_location();
  return err;
}

int prefetch_destroy()
{
  int err = 0;

  print_entry_location();

  print_exit_location();
  return err;
}

static int _request_page(struct file* filp,int offset,int length)
{
  int err = 0;
  page_t *page;

  print_entry_location();

  page = read_cache_page(filp->f_dentry->d_inode->i_mapping,
		  offset >> PAGE_CACHE_SHIFT,
		  (filler_t*)filp->f_dentry->d_inode->i_mapping->a_ops->readpage,
		  NULL);

  //Should never span two pages - the UL compiler handles this
  ASSERT(length + (offset % PAGE_SIZE) <= PAGE_SIZE);

  print_exit_location();
  return err;
}

/*
 * This decreases the window size linearly.
 * This is used when there are no misses in the readahead
 */
static inline int _decrease_window_size(void)
{
  if(prefetch_window_size - WINDOW_SHRINK > 0)
    prefetch_window_size -= WINDOW_SHRINK;
  return prefetch_window_size;
}

/*
 * This increases the window size exponentially
 * This is used when there are misses in the readahead
 */
static inline int _increase_window_size(void)
{
  return prefetch_window_size *= WINDOW_EXPAND;
}

//Decides whether readahead should continue
static inline int _cmd_based_readahead(struct replay *r)
{
  return (r->cur_cmd < r->cmd_num && (last_rcmd_num - r->cur_cmd) < WINDOW_CMD_SIZE);
}

//Decides whether readahead should continue
static int _time_based_readahead(struct replay *r)
{
  int ret = 0;
  long ms_diff = 0;

  if(r->cur_cmd >= r->cmd_num || last_rcmd_num >= r->cmd_num) {
    ret = 0;
    goto out;
  }

  if(!last_rcmd_num) {
    ret = 1;
    goto out;
  }

  ms_diff = (r->srcmdp[last_rcmd_num].time.secs * 1000) - (r->srcmdp[r->cur_cmd].time.secs * 1000);
  ms_diff += (r->srcmdp[last_rcmd_num].time.usecs / 1000) - (r->srcmdp[r->cur_cmd].time.usecs / 1000);

  if(ms_diff < prefetch_window_size)
    ret = 1;

 out:
  return ret;
}

static int _readahead(struct replay* r,struct file* filp)
{
  int err = 0;
  print_entry_location();

  while(TIME_BASED_WINDOW_SIZE ? _time_based_readahead(r) : _cmd_based_readahead(r)) {
    if(r->srcmdp[last_rcmd_num].command == OP_LOAD)
      _request_page(filp,r->srcmdp[last_rcmd_num].args[0],r->srcmdp[last_rcmd_num].args[1]);
    last_rcmd_num++;
  }

  //Decrease Window Size
  _decrease_window_size();

  print_exit_location();
  return err;
}

int prefetch_readahead(struct replay *r,struct file* filp)
{
  int err = 0;

  print_entry_location();

  //Perform Readahead
  _readahead(r,filp);

  print_exit_location();
  return err;
}

//Returns the buffer - assume _request_page() was previously called
void* prefetch_get_buffer(struct file* filp,long offset,long length)
{
  void* ret;
  page_t* page;
  char* page_data = NULL;

  print_entry_location();
  ASSERT(length <= PAGE_SIZE); //Should not get requests larger than a page

  page = find_get_page(filp->f_dentry->d_inode->i_mapping,(int)(offset >> PAGE_CACHE_SHIFT));

  if(!page) {
    //BUG, means request_page() wasn't called!
    print_error_message("BUG: request_page() was not called before!");
    _request_page(filp,offset,0);
    goto wait;
  }

  //If page readahead missed!  Increase the window size
  if(!PageUptodate(page))
    _increase_window_size();

 wait:
  //Wait until the page is unlocked
  wait_on_page_locked(page);

  while(!PageUptodate(page))
    schedule(); //Wait

  ASSERT(length + (offset % PAGE_SIZE) <= PAGE_SIZE);

  //Map the data in the page
  page_data = kmap(page);
  ret = page_data + (offset % PAGE_SIZE);

  //Decrement ref count - incremented in find_get_page()
  page_cache_release(page); //for find_get_page()

 out:
  print_exit_location();
  return ret;
}

int prefetch_release_buffer(struct file* filp,void* data,long offset)
{
  int err = 0;
  page_t* page;

  print_entry_location();
  page = find_get_page(filp->f_dentry->d_inode->i_mapping,offset >> PAGE_CACHE_SHIFT);
  ASSERT(page);
  kunmap(page);
  page_cache_release(page); //for find_get_page()
  page_cache_release(page); //for read_cache_page()

 out:
  print_exit_location();
  return err;
}
